/* Based on K8 decoding code written for the 2.4 kernel by Andi Kleen and 
 * Eric Morton. Hacked and extended for mcelog by AK.
 * Extended to support all AMD families by Borislav Petkov, SUSE Labs.
 *
 * Original copyright: 
 * K8 parts Copyright 2002,2003 Andi Kleen, SuSE Labs.
 * Additional K8 decoding and simplification Copyright 2003 Eric Morton, Newisys Inc 
 * K8 threshold counters decoding Copyright 2005,2006 Jacob Shin, AMD Inc.
 * 
 * Subject to the GNU General Public License
 */

#include <stdio.h>
#include "mcelog.h"
#include "amd.h"

static const char * const k8bank[] = {
	"data cache",
	"instruction cache",
	"bus unit",
	"load/store unit",
	"northbridge",
	"fixed-issue reoder"
};
static const char * const transaction[] = {
	"instruction", "data", "generic", "reserved"
};
static const char * const cachelevel[] = {
	"0", "1", "2", "generic"
};
static const char * const memtrans[] = {
	"generic error", "generic read", "generic write", "data read",
	"data write", "instruction fetch", "prefetch", "evict", "snoop",
	"?", "?", "?", "?", "?", "?", "?"
};
static const char * const partproc[] = {
	"local node origin",
	"local node response",
	"local node observed",
	"generic participation"
};
static const char * const timeout[] = {
	"request didn't time out",
	"request timed out"
};
static const char * const memoryio[] = {
	"memory", "res.", "i/o", "generic"
};

/* internal error type */
static const char * const uu_msgs[] = { "RESV", "RESV", "HWA", "RESV" };

static char *nbextendederr[] = { 
	"RAM ECC error", 
	"CRC error",
	"Sync error",
	"Master abort",
	"Target abort",
	"GART error",
	"RMW error",
	"Watchdog error",
	"RAM Chipkill ECC error", 
	"DEV Error",
	"Link Data Error",
	"Link Protocol Error",
	"NB Array Error",
	"DRAM Parity Error",
	"Link Retry",
	"Tablew Walk Data Error",
	"L3 Cache Data Error",
	"L3 Cache Tag Error",
	"L3 Cache LRU Error"
};

static const char * const f15h_mc1_mce_desc[] = {
	"UC during a demand linefill from L2",
	"Parity error during data load from IC",
	"Parity error for IC valid bit",
	"Main tag parity error",
	"Parity error in prediction queue",
	"PFB data/address parity error",
	"Parity error in the branch status reg",
	"PFB promotion address error",
	"Tag error during probe/victimization",
	"Parity error for IC probe tag valid bit",
	"PFB non-cacheable bit parity error",
	"PFB valid bit parity error",			/* xec = 0xd */
	"Microcode Patch Buffer",			/* xec = 010 */
	"uop queue",
	"insn buffer",
	"predecode buffer",
	"fetch address FIFO"
};

static const char * const f15h_mc2_mce_desc[] = {
	"Fill ECC error on data fills",			/* xec = 0x4 */
	"Fill parity error on insn fills",
	"Prefetcher request FIFO parity error",
	"PRQ address parity error",
	"PRQ data parity error",
	"WCC Tag ECC error",
	"WCC Data ECC error",
	"WCB Data parity error",
	"VB Data ECC or parity error",
	"L2 Tag ECC error",				/* xec = 0x10 */
	"Hard L2 Tag ECC error",
	"Multiple hits on L2 tag",
	"XAB parity error",
	"PRB address parity error"
};

static const char * const mc4_mce_desc[] = {
	"DRAM ECC error detected on the NB",
	"CRC error detected on HT link",
	"Link-defined sync error packets detected on HT link",
	"HT Master abort",
	"HT Target abort",
	"Invalid GART PTE entry during GART table walk",
	"Unsupported atomic RMW received from an IO link",
	"Watchdog timeout due to lack of progress",
	"DRAM ECC error detected on the NB",
	"SVM DMA Exclusion Vector error",
	"HT data error detected on link",
	"Protocol error (link, L3, probe filter)",
	"NB internal arrays parity error",
	"DRAM addr/ctl signals parity error",
	"IO link transmission error",
	"L3 data cache ECC error",			/* xec = 0x1c */
	"L3 cache tag error",
	"L3 LRU parity bits error",
	"ECC Error in the Probe Filter directory"
};

static const char * const mc5_mce_desc[] = {
	"CPU Watchdog timer expire",
	"Wakeup array dest tag",
	"AG payload array",
	"EX payload array",
	"IDRF array",
	"Retire dispatch queue",
	"Mapper checkpoint array",
	"Physical register file EX0 port",
	"Physical register file EX1 port",
	"Physical register file AG0 port",
	"Physical register file AG1 port",
	"Flag register file",
	"DE error occurred",
	"Retire status queue"
};

static char *highbits[32] = { 
	[31] = "valid",
	[30] = "error overflow (multiple errors)",
	[29] = "error uncorrected",
	[28] = "error enable",
	[27] = "misc error valid",
	[26] = "error address valid", 
	[25] = "processor context corrupt", 
	[24] = "res24",
	[23] = "res23",
	/* 22-15 ecc syndrome bits */
	[14] = "corrected ecc error",
	[13] = "uncorrected ecc error",
	[12] = "res12",
	[11] = "L3 subcache in error bit 1",
	[10] = "L3 subcache in error bit 0",
	[9] = "sublink or DRAM channel",
	[8] = "error found by scrub", 
	/* 7-4 ht link number of error */ 
	[3] = "err cpu3",
	[2] = "err cpu2",
	[1] = "err cpu1",
	[0] = "err cpu0",
};
static char *k8threshold[] = {
	[0 ... K8_MCELOG_THRESHOLD_DRAM_ECC - 1] = "Unknow threshold counter",
	[K8_MCELOG_THRESHOLD_DRAM_ECC] = "MC4_MISC0 DRAM threshold",
	[K8_MCELOG_THRESHOLD_LINK] = "MC4_MISC1 Link threshold",
	[K8_MCELOG_THRESHOLD_L3_CACHE] = "MC4_MISC2 L3 Cache threshold",
	[K8_MCELOG_THRESHOLD_FBDIMM] = "MC4_MISC3 FBDIMM threshold",
	[K8_MCELOG_THRESHOLD_FBDIMM + 1 ... 
	 K8_MCE_THRESHOLD_TOP - K8_MCE_THRESHOLD_BASE - 1] = 
                "Unknown threshold counter",
};

static u8 xec_mask = 0xf;

enum cputype select_amd_cputype(u32 family)
{
	switch (family) {
	case 0xf:
		return CPU_K8;
	case 0x10:
		return CPU_F10H;
	case 0x11:
		return CPU_F11H;
	case 0x12:
		return CPU_F12H;
	case 0x14:
		return CPU_F14H;
	case 0x15:
		return CPU_F15H;
	case 0x16:
		return CPU_F16H;
	default:
		break;
	}

	return CPU_GENERIC;
}

static void decode_k8_generic_errcode(u64 status)
{
	unsigned short errcode = status & 0xffff;
	int i;

	for (i=0; i<32; i++) {
		if (i==31 || i==28 || i==26)
			continue;
		if (highbits[i] && (status & (1ULL<<(i+32)))) {
			Wprintf( "       bit%d = %s\n", i+32, highbits[i]);
		}
	}

	if ((errcode & 0xFFF0) == 0x0010) {
		Wprintf( "  TLB error '%s transaction, level %s'\n",
		       transaction[(errcode >> 2) & 3],
		       cachelevel[errcode & 3]);
	}
	else if ((errcode & 0xFF00) == 0x0100) {
		Wprintf( "  memory/cache error '%s mem transaction, %s transaction, level %s'\n",
		       memtrans[(errcode >> 4) & 0xf],
		       transaction[(errcode >> 2) & 3],
		       cachelevel[errcode & 3]);
	}
	else if ((errcode & 0xF800) == 0x0800) {
		Wprintf( "  bus error '%s, %s\n             %s mem transaction\n             %s access, level %s'\n",
		       partproc[(errcode >> 9) & 0x3],
		       timeout[(errcode >> 8) & 1],
		       memtrans[(errcode >> 4) & 0xf],
		       memoryio[(errcode >> 2) & 0x3],
		       cachelevel[(errcode & 0x3)]);
	}
}

static void decode_k8_dc_mc(u64 status, int *err)
{
	unsigned short exterrcode = (status >> 16) & 0x0f;
	unsigned short errcode = status & 0xffff;

	if(status&(3ULL<<45)) {
		Wprintf( "  Data cache ECC error (syndrome %x)",
		       (u32) (status >> 47) & 0xff);
		if(status&(1ULL<<40)) {
			Wprintf(" found by scrubber");
		}
		Wprintf("\n");
	}

	if ((errcode & 0xFFF0) == 0x0010) {
		Wprintf( "  TLB parity error in %s array\n",
		       (exterrcode == 0) ? "physical" : "virtual");
	}

	decode_k8_generic_errcode(status);
}

static void decode_k8_ic_mc(u64 status, int *err)
{
	unsigned short exterrcode = (status >> 16) & 0x0f;
	unsigned short errcode = status & 0xffff;

	if(status&(3ULL<<45)) {
		Wprintf("  Instruction cache ECC error\n");
	}

	if ((errcode & 0xFFF0) == 0x0010) {
		Wprintf("  TLB parity error in %s array\n",
		       (exterrcode == 0) ? "physical" : "virtual");
	}

	decode_k8_generic_errcode(status);
}

static void decode_k8_bu_mc(u64 status, int *err)
{
	unsigned short exterrcode = (status >> 16) & 0x0f;

	if(status&(3ULL<<45)) {
		Wprintf("  L2 cache ECC error\n");
	}

	Wprintf("  %s array error\n",
	       (exterrcode == 0) ? "Bus or cache" : "Cache tag");

	decode_k8_generic_errcode(status);
}

static void decode_k8_ls_mc(u64 status, int *err)
{
	decode_k8_generic_errcode(status);
}

static void decode_k8_nb_mc(u64 status, int *memerr)
{
	unsigned short exterrcode = (status >> 16) & 0x0f;

	Wprintf("  Northbridge %s\n", nbextendederr[exterrcode]);

	switch (exterrcode) { 
	case 0:
		*memerr = 1;
		Wprintf("  ECC syndrome = %x\n",
		       (u32) (status >> 47) & 0xff);
		break;
	case 8:	
		*memerr = 1;
		Wprintf("  Chipkill ECC syndrome = %x\n",
		       (u32) ((((status >> 24) & 0xff) << 8) | ((status >> 47) & 0xff)));
		break;
	case 1: 
	case 2:
	case 3:
	case 4:
	case 6:
		Wprintf("  link number = %x\n",
		       (u32) (status >> 36) & 0xf);
		break;		   
	}

	decode_k8_generic_errcode(status);
}

static void decode_k8_fr_mc(u64 status, int *err)
{
	decode_k8_generic_errcode(status);
}

static void decode_k8_threshold(u64 misc)
{	
	if (misc & MCI_THRESHOLD_OVER)
		Wprintf("  Threshold error count overflow\n");
}

typedef void (*decoder_t)(u64, int *ismemerr); 

static decoder_t decoders[] = {
	[0] = decode_k8_dc_mc,
	[1] = decode_k8_ic_mc,
	[2] = decode_k8_bu_mc,
	[3] = decode_k8_ls_mc,
	[4] = decode_k8_nb_mc,
	[5] = decode_k8_fr_mc,
};

static bool k8_mc1_mce(u16 ec, u8 xec)
{
	u8 ll	 = LL(ec);
	bool ret = true;

	if (!MEM_ERROR(ec))
		return false;

	if (ll == 0x2)
		Wprintf("during a linefill from L2.\n");
	else if (ll == 0x1) {
		switch (R4(ec)) {
		case R4_IRD:
			Wprintf("Parity error during data load.\n");
			break;

		case R4_EVICT:
			Wprintf("Copyback Parity/Victim error.\n");
			break;

		case R4_SNOOP:
			Wprintf("Tag Snoop error.\n");
			break;

		default:
			ret = false;
			break;
		}
	} else
		ret = false;

	return ret;
}

static bool f12h_mc0_mce(u16 ec, u8 xec)
{
	bool ret = false;

	if (MEM_ERROR(ec)) {
		u8 ll = LL(ec);
		ret = true;

		if (ll == LL_L2)
			Wprintf("aduring L1 linefill from L2.\n");
		else if (ll == LL_L1)
			Wprintf("Data/Tag %s error.\n", R4_MSG(ec));
		else
			ret = false;
	}
	return ret;
}

static bool f10h_mc0_mce(u16 ec, u8 xec)
{
	if (R4(ec) == R4_GEN && LL(ec) == LL_L1) {
		Wprintf("during data scrub.\n");
		return true;
	}
	return f12h_mc0_mce(ec, xec);
}

static bool k8_mc0_mce(u16 ec, u8 xec)
{
	if (BUS_ERROR(ec)) {
		Wprintf("during system linefill.\n");
		return true;
	}

	return f10h_mc0_mce(ec, xec);
}

static bool cat_mc0_mce(u16 ec, u8 xec)
{
	u8 r4	 = R4(ec);
	bool ret = true;

	if (MEM_ERROR(ec)) {

		if (TT(ec) != TT_DATA || LL(ec) != LL_L1)
			return false;

		switch (r4) {
		case R4_DRD:
		case R4_DWR:
			Wprintf("Data/Tag parity error due to %s.\n",
				(r4 == R4_DRD ? "load/hw prf" : "store"));
			break;
		case R4_EVICT:
			Wprintf("Copyback parity error on a tag miss.\n");
			break;
		case R4_SNOOP:
			Wprintf("Tag parity error during snoop.\n");
			break;
		default:
			ret = false;
		}
	} else if (BUS_ERROR(ec)) {

		if ((II(ec) != II_MEM && II(ec) != II_IO) || LL(ec) != LL_LG)
			return false;

		Wprintf("System read data error on a ");

		switch (r4) {
		case R4_RD:
			Wprintf("TLB reload.\n");
			break;
		case R4_DWR:
			Wprintf("store.\n");
			break;
		case R4_DRD:
			Wprintf("load.\n");
			break;
		default:
			ret = false;
		}
	} else {
		ret = false;
	}

	return ret;
}

static bool f15h_mc0_mce(u16 ec, u8 xec)
{
	bool ret = true;

	if (MEM_ERROR(ec)) {

		switch (xec) {
		case 0x0:
			Wprintf("Data Array access error.\n");
			break;

		case 0x1:
			Wprintf("UC error during a linefill from L2/NB.\n");
			break;

		case 0x2:
		case 0x11:
			Wprintf("STQ access error.\n");
			break;

		case 0x3:
			Wprintf("SCB access error.\n");
			break;

		case 0x10:
			Wprintf("Tag error.\n");
			break;

		case 0x12:
			Wprintf("LDQ access error.\n");
			break;

		default:
			ret = false;
		}
	} else if (BUS_ERROR(ec)) {

		if (!xec)
			Wprintf("System Read Data Error.\n");
		else
			Wprintf(" Internal error condition type %d.\n", xec);
	} else
		ret = false;

	return ret;
}

static void decode_mc0_mce(struct amd_decoder_ops *ops, struct mce *m)
{
	u16 ec = EC(m->status);
	u8 xec = XEC(m->status, xec_mask);

	Wprintf("  MC0 Error: ");

	/* TLB error signatures are the same across families */
	if (TLB_ERROR(ec)) {
		if (TT(ec) == TT_DATA) {
			Wprintf("%s TLB %s.\n", LL_MSG(ec),
				((xec == 2) ? "locked miss"
					    : (xec ? "multimatch" : "parity")));
			return;
		}
	} else if (ops->mc0_mce(ec, xec))
		;
	else
		Eprintf("Corrupted MC0 MCE info?\n");
}

static bool cat_mc1_mce(u16 ec, u8 xec)
{
	u8 r4    = R4(ec);
	bool ret = true;

	if (!MEM_ERROR(ec))
		return false;

	if (TT(ec) != TT_INSTR)
		return false;

	if (r4 == R4_IRD)
		Wprintf("Data/tag array parity error for a tag hit.\n");
	else if (r4 == R4_SNOOP)
		Wprintf("Tag error during snoop/victimization.\n");
	else if (xec == 0x0)
		Wprintf("Tag parity error from victim castout.\n");
	else if (xec == 0x2)
		Wprintf("Microcode patch RAM parity error.\n");
	else
		ret = false;

	return ret;
}

static bool f15h_mc1_mce(u16 ec, u8 xec)
{
	bool ret = true;

	if (!MEM_ERROR(ec))
		return false;

	switch (xec) {
	case 0x0 ... 0xa:
		Wprintf("%s.\n", f15h_mc1_mce_desc[xec]);
		break;

	case 0xd:
		Wprintf("%s.\n", f15h_mc1_mce_desc[xec-2]);
		break;

	case 0x10:
		Wprintf("%s.\n", f15h_mc1_mce_desc[xec-4]);
		break;

	case 0x11 ... 0x14:
		Wprintf("Decoder %s parity error.\n", f15h_mc1_mce_desc[xec-4]);
		break;

	default:
		ret = false;
	}
	return ret;
}

static void decode_mc1_mce(struct amd_decoder_ops *ops, struct mce *m)
{
	u16 ec = EC(m->status);
	u8 xec = XEC(m->status, xec_mask);

	Wprintf("  MC1 Error: ");

	if (TLB_ERROR(ec))
		Wprintf("%s TLB %s.\n", LL_MSG(ec),
			(xec ? "multimatch" : "parity error"));
	else if (BUS_ERROR(ec)) {
		bool k8 = ((ops->cpu == AMD_K8) && (m->status & BIT_64(58)));

		Wprintf("during %s.\n", (k8 ? "system linefill" : "NB data read"));
	} else if (ops->mc1_mce(ec, xec))
		;
	else
		Eprintf("Corrupted MC1 MCE info?\n");
}

static bool k8_mc2_mce(u16 ec, u8 xec)
{
	bool ret = true;

	if (xec == 0x1)
		Wprintf(" in the write data buffers.\n");
	else if (xec == 0x3)
		Wprintf(" in the victim data buffers.\n");
	else if (xec == 0x2 && MEM_ERROR(ec))
		Wprintf(": %s error in the L2 cache tags.\n", R4_MSG(ec));
	else if (xec == 0x0) {
		if (TLB_ERROR(ec))
			Wprintf(": %s error in a Page Descriptor Cache or "
				"Guest TLB.\n", TT_MSG(ec));
		else if (BUS_ERROR(ec))
			Wprintf(": %s/ECC error in data read from NB: %s.\n",
				R4_MSG(ec), PP_MSG(ec));
		else if (MEM_ERROR(ec)) {
			u8 r4 = R4(ec);

			if (r4 >= 0x7)
				Wprintf(": %s error during data copyback.\n",
					R4_MSG(ec));
			else if (r4 <= 0x1)
				Wprintf(": %s parity/ECC error during data "
					"access from L2.\n", R4_MSG(ec));
			else
				ret = false;
		} else
			ret = false;
	} else
		ret = false;

	return ret;
}

static bool f15h_mc2_mce(u16 ec, u8 xec)
{
	bool ret = true;

	if (TLB_ERROR(ec)) {
		if (xec == 0x0)
			Wprintf("Data parity TLB read error.\n");
		else if (xec == 0x1)
			Wprintf("Poison data provided for TLB fill.\n");
		else
			ret = false;
	} else if (BUS_ERROR(ec)) {
		if (xec > 2)
			ret = false;

		Wprintf("Error during attempted NB data read.\n");
	} else if (MEM_ERROR(ec)) {
		switch (xec) {
		case 0x4 ... 0xc:
			Wprintf("%s.\n", f15h_mc2_mce_desc[xec - 0x4]);
			break;

		case 0x10 ... 0x14:
			Wprintf("%s.\n", f15h_mc2_mce_desc[xec - 0x7]);
			break;

		default:
			ret = false;
		}
	}

	return ret;
}

static bool f16h_mc2_mce(u16 ec, u8 xec)
{
	u8 r4 = R4(ec);

	if (!MEM_ERROR(ec))
		return false;

	switch (xec) {
	case 0x04 ... 0x05:
		Wprintf("%cBUFF parity error.\n", (r4 == R4_RD) ? 'I' : 'O');
		break;

	case 0x09 ... 0x0b:
	case 0x0d ... 0x0f:
		Wprintf("ECC error in L2 tag (%s).\n",
			((r4 == R4_GEN)   ? "BankReq" :
			((r4 == R4_SNOOP) ? "Prb"     : "Fill")));
		break;

	case 0x10 ... 0x19:
	case 0x1b:
		Wprintf("ECC error in L2 data array (%s).\n",
			(((r4 == R4_RD) && !(xec & 0x3)) ? "Hit"  :
			((r4 == R4_GEN)   ? "Attr" :
			((r4 == R4_EVICT) ? "Vict" : "Fill"))));
		break;

	case 0x1c ... 0x1d:
	case 0x1f:
		Wprintf("Parity error in L2 attribute bits (%s).\n",
			((r4 == R4_RD)  ? "Hit"  :
			((r4 == R4_GEN) ? "Attr" : "Fill")));
		break;

	default:
		return false;
	}

	return true;
}

static void decode_mc2_mce(struct amd_decoder_ops *ops, struct mce *m)
{
	u16 ec = EC(m->status);
	u8 xec = XEC(m->status, xec_mask);

	Wprintf("  MC2 Error: ");

	if (!ops->mc2_mce(ec, xec))
		Eprintf("Corrupted MC2 MCE info?\n");
}

static void decode_mc3_mce(struct amd_decoder_ops *ops, struct mce *m)
{
	u16 ec = EC(m->status);
	u8 xec = XEC(m->status, xec_mask);

	if (ops->cpu >= AMD_F14H) {
		Eprintf("You shouldn't be seeing MC3 MCE on this cpu family,"
			 " please report on LKML.\n");
		return;
	}

	Wprintf("  MC3 Error");

	if (xec == 0x0) {
		u8 r4 = R4(ec);

		if (!BUS_ERROR(ec) || (r4 != R4_DRD && r4 != R4_DWR))
			goto wrong_mc3_mce;

		Wprintf(" during %s.\n", R4_MSG(ec));
	} else
		goto wrong_mc3_mce;

	return;

wrong_mc3_mce:
	Eprintf("Corrupted MC3 MCE info?\n");
}

static void decode_mc4_mce(struct amd_decoder_ops *ops, struct mce *m)
{
	u16 ec = EC(m->status);
	u8 xec = XEC(m->status, 0x1f);
	u8 offset = 0;

	Wprintf("  MC4 Error: ");

	switch (xec) {
	case 0x0 ... 0xe:

		/* special handling for DRAM ECCs */
		if (xec == 0x0 || xec == 0x8) {
			/* no ECCs on F11h */
			if (ops->cpu == AMD_F11H)
				goto wrong_mc4_mce;

			Wprintf("%s.\n", mc4_mce_desc[xec]);
			return;
		}
		break;

	case 0xf:
		if (TLB_ERROR(ec))
			Wprintf("GART Table Walk data error.\n");
		else if (BUS_ERROR(ec))
			Wprintf("DMA Exclusion Vector Table Walk error.\n");
		else
			goto wrong_mc4_mce;
		return;

	case 0x19:
		if (ops->cpu >= AMD_F15H || ops->cpu <= AMD_F16H)
			Wprintf("Compute Unit Data Error.\n");
		else
			goto wrong_mc4_mce;
		return;

	case 0x1c ... 0x1f:
		offset = 13;
		break;

	default:
		goto wrong_mc4_mce;
	}

	Wprintf("%s.\n", mc4_mce_desc[xec - offset]);
	return;

 wrong_mc4_mce:
	Eprintf("Corrupted MC4 MCE info?\n");
}

static void decode_mc5_mce(struct amd_decoder_ops *ops, struct mce *m)
{
	u8 xec = XEC(m->status, xec_mask);

	if (ops->cpu == AMD_K8 || ops->cpu == AMD_F11H)
		goto wrong_mc5_mce;

	Wprintf("  MC5 Error: ");

	if (xec == 0x0 || xec == 0xc)
		Wprintf("%s.\n", mc5_mce_desc[xec]);
	else if (xec <= 0xd)
		Wprintf("%s parity error.\n", mc5_mce_desc[xec]);
	else
		goto wrong_mc5_mce;

	return;

 wrong_mc5_mce:
	Eprintf("Corrupted MC5 MCE info?\n");
}

static void decode_mc6_mce(struct mce *m)
{
	u8 xec = XEC(m->status, xec_mask);

	Wprintf("  MC6 Error: ");

	switch (xec) {
	case 0x1:
		Wprintf("Free List");
		break;

	case 0x2:
		Wprintf("Physical Register File");
		break;

	case 0x3:
		Wprintf("Retire Queue");
		break;

	case 0x4:
		Wprintf("Scheduler table");
		break;

	case 0x5:
		Wprintf("Status Register File");
		break;

	default:
		goto wrong_mc6_mce;
		break;
	}

	Wprintf(" parity error.\n");

	return;

 wrong_mc6_mce:
	Eprintf("Corrupted MC6 MCE info?\n");
}

static inline void amd_decode_err_code(u16 ec)
{
	if (INT_ERROR(ec)) {
		Wprintf("  internal: %s\n", UU_MSG(ec));
		return;
	}

	Wprintf("  cache level: %s", LL_MSG(ec));

	if (BUS_ERROR(ec))
		Wprintf(", mem/io: %s", II_MSG(ec));
	else
		Wprintf(", tx: %s", TT_MSG(ec));

	if (MEM_ERROR(ec) || BUS_ERROR(ec)) {
		Wprintf(", mem-tx: %s", R4_MSG(ec));

		if (BUS_ERROR(ec))
			Wprintf(", part-proc: %s (%s)", PP_MSG(ec), TO_MSG(ec));
	}

	Wprintf("\n");
}

struct amd_decoder_ops fam_ops[] = {
	[AMD_F10H] = {
		.cpu	 = AMD_F10H,
		.mc0_mce = f10h_mc0_mce,
		.mc1_mce = k8_mc1_mce,
		.mc2_mce = k8_mc2_mce,
	},
	[AMD_F11H] = {
		.cpu	 = AMD_F11H,
		.mc0_mce = k8_mc0_mce,
		.mc1_mce = k8_mc1_mce,
		.mc2_mce = k8_mc2_mce,
	},
	[AMD_F12H] = {
		.cpu	 = AMD_F12H,
		.mc0_mce = f12h_mc0_mce,
		.mc1_mce = k8_mc1_mce,
		.mc2_mce = k8_mc2_mce,
	},
	[AMD_F14H] = {
		.cpu	 = AMD_F14H,
		.mc0_mce = cat_mc0_mce,
		.mc1_mce = cat_mc1_mce,
		.mc2_mce = k8_mc2_mce,
	},
	[AMD_F15H] = {
		.cpu = AMD_F15H,
		.mc0_mce = f15h_mc0_mce,
		.mc1_mce = f15h_mc1_mce,
		.mc2_mce = f15h_mc2_mce,
	},
	[AMD_F16H] = {
		.cpu = AMD_F16H,
		.mc0_mce = cat_mc0_mce,
		.mc1_mce = cat_mc1_mce,
		.mc2_mce = f16h_mc2_mce,
	},
};

static void __decode_amd_mc(enum cputype cpu, struct mce *mce)
{
	struct amd_decoder_ops *ops;

	switch (cpu) {
	case CPU_F10H:
		ops = &fam_ops[AMD_F10H];
		break;
	case CPU_F11H:
		ops = &fam_ops[AMD_F11H];
		break;
	case CPU_F12H:
		ops = &fam_ops[AMD_F12H];
		break;
	case CPU_F14H:
		ops = &fam_ops[AMD_F14H];
		break;
	case CPU_F15H:
		xec_mask = 0x1f;
		ops = &fam_ops[AMD_F15H];
		break;
	case CPU_F16H:
		xec_mask = 0x1f;
		ops = &fam_ops[AMD_F16H];
		break;
	default:
		Eprintf("Huh? What family is it: 0x%x?!\n", cpu);
		return;
		break;
	}

	switch (mce->bank) {
	case 0:
		decode_mc0_mce(ops, mce);
		break;
	case 1:
		decode_mc1_mce(ops, mce);
		break;
	case 2:
		decode_mc2_mce(ops, mce);
		break;
	case 3:
		decode_mc3_mce(ops, mce);
		break;
	case 4:
		decode_mc4_mce(ops, mce);
		break;
	case 5:
		decode_mc5_mce(ops, mce);
		break;
	case 6:
		decode_mc6_mce(mce);
		break;

	default:
		break;
	}
	amd_decode_err_code(mce->status & 0xffff);
}

void decode_amd_mc(enum cputype cpu, struct mce *mce, int *ismemerr)
{
	if (cpu == CPU_K8) {
		if (mce->bank < NELE(decoders))
			decoders[mce->bank](mce->status, ismemerr);
		else if (mce->bank >= K8_MCE_THRESHOLD_BASE &&
			 mce->bank < K8_MCE_THRESHOLD_TOP)
			decode_k8_threshold(mce->misc);
		else
			Wprintf("  no decoder for unknown bank %u\n", mce->bank);
	} else
		__decode_amd_mc(cpu, mce);
}

char *k8_bank_name(unsigned num)
{ 
	static char buf[64];
	const char *s = "unknown";
	if (num < NELE(k8bank))
		s = k8bank[num];
	else if (num >= K8_MCE_THRESHOLD_BASE && 
		 num < K8_MCE_THRESHOLD_TOP)
		s = k8threshold[num - K8_MCE_THRESHOLD_BASE];
	buf[sizeof(buf)-1] = 0;
	snprintf(buf, sizeof(buf) - 1, "%u %s", num, s);
	return buf;
}

int mce_filter_amd(struct mce *m)
{
        /*
         * NB GART TLB error reporting is disabled by default.
         */
	if (m->bank == 4) {
		u8 xec = (m->status >> 16) & 0x1f;

		if (xec == 0x5 && (m->status & BIT_64(61)))
			return 0;
	}
        return 1;
}
