#include <unistd.h>
#include <signal.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define __USE_GNU
/* To fetch the dnsname */
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include <libesmtp.h>
#include "mcelog.h"
#include "email.h"

#define MAX_STRING_LEN 512
char c_recipient[MAX_STRING_LEN] = "";
static int debug;
static char dnsname[MAX_STRING_LEN];

static char buf[128];
#define ERROR() { fprintf (stderr, "SMTP problem [%d] %s\n", __LINE__,	\
			 smtp_strerror (smtp_errno (), buf, sizeof buf));	\
		return -1; }


void email_usage(void) {
	fprintf(stderr,
		"--email address     Requires daemon mode\n");
}

int email_cmd(int opt, int ac, char **av)
{
	char *arg = optarg;

	switch (opt) {
	case O_EMAIL_ADDRESS:
		if (arg) {
			if (strlen(arg) >= MAX_STRING_LEN) {
				Eprintf("email address too long"
					" [max:%d]\n", MAX_STRING_LEN);
				return 0;
			}
			strcpy(c_recipient, arg);
			return 1;
		}
	case O_EMAIL_DEBUG:
		debug = 1;
		return 0;
	}
	return 0;
}

int email_env(void)
{
	char *email_env = getenv("MCELOG_EMAIL_DEBUG");

	if (email_env)
		debug=0;

	email_env = getenv("MCELOG_ADMIN_EMAIL");
	/* No email validation, but at least check for not being empty... */
	if (email_env && strlen(email_env) > 1) {
		strncpy(c_recipient, email_env, MAX_STRING_LEN - 1);
		return 1;
	}
	return 0;
}

/* Callback to prnt the recipient status */
static void
print_recipient_status (smtp_recipient_t recipient,
                        const char *mailbox, void *arg)
{
	const smtp_status_t *status;

	status = smtp_recipient_status (recipient);
	if (debug)
		printf ("%s: %d %s", mailbox, status->code, status->text);
}

void setup_mail_header(FILE *fp, struct mce *m)
{
	char host[MAX_STRING_LEN];
	struct addrinfo hints;
	struct addrinfo *res=NULL;
	int    ret, retry=3;

	/* Taken from net-tools hostname.c showhname() */
	memset(&hints, 0, sizeof(struct addrinfo));
	hints.ai_family   = AF_UNSPEC;
	hints.ai_flags    = AI_CANONNAME | AI_CANONIDN;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = 0;

	if (gethostname(host, MAX_STRING_LEN)) {
		fprintf(stderr, "Cannot get host name\n");
		return;
	}

	do {
		ret = getaddrinfo(host, NULL, &hints, &res);
	} while(ret == EAI_AGAIN && retry-- > 0
                             && usleep(50000) == 0);

	if (ret != 0 || res == NULL) {
		fprintf(stderr, "Could not retrieve hostname\n");
		return;
	}

	memset(dnsname, '\0', MAX_STRING_LEN);
	strncpy(dnsname, res->ai_canonname, MAX_STRING_LEN - 1);

	fprintf(fp, "Return-Path: <dummy@will_get_overridden.net>\r\n"
		"Subject: Machine Check Exception on %s detected\r\n"
		"MIME-Version: 1.0\r\n"
		"Content-Type: text/plain;\r\n"
		"  charset=iso-8859-1\r\n"
		"Content-Transfer-Encoding: 7bit\r\n\r\n", dnsname);
	freeaddrinfo(res);
}


int send_mail(FILE *fp)
{
	char smtp_host[MAX_STRING_LEN] = "localhost:25";
	char from[MAX_STRING_LEN];

	const smtp_status_t *status;
	smtp_session_t session;
	smtp_message_t message;
	smtp_recipient_t recipient;
	struct sigaction sa;

	session = smtp_create_session ();
	message = smtp_add_message (session);

	snprintf(from, MAX_STRING_LEN, "root@%s", dnsname);

	/* NB.  libESMTP sets timeouts as it progresses through the protocol.
	   In addition the remote server might close its socket on a timeout.
	   Consequently libESMTP may sometimes try to write to a socket with
	   no reader.  Ignore SIGPIPE, then the program doesn't get killed
	   if/when this happens. */
	sa.sa_handler = SIG_IGN;
	sigemptyset (&sa.sa_mask);
	sa.sa_flags = 0;
	sigaction (SIGPIPE, &sa, NULL);

	/* Set the host running the SMTP server.  LibESMTP has a default port
	   number of 587, however this is not widely deployed so the port
	   is specified as 25 along with the default MTA host. */
	if (!smtp_set_server (session, smtp_host))
		ERROR();

	smtp_set_reverse_path (message, from);

	/* RFC 2822 doesn't require recipient headers but a To: header would
	   be nice to have if not present. */
	smtp_set_header (message, "To", NULL, NULL);

	/* RFC 2822 doesn't require recipient headers but a To: header would
	   be nice to have if not present. */
	if (!smtp_set_header (message, "From", "mcelog", from))
		ERROR();

	smtp_set_message_fp (message, fp);

	recipient = smtp_add_recipient (message, c_recipient);
	if (!recipient)
		ERROR();
	if (!smtp_dsn_set_notify (recipient, Notify_NEVER))
		ERROR();

	/* Initiate a connection to the SMTP server and transfer the
	   message. */
	if (!smtp_start_session (session))
		Eprintf("SMTP server problem %s\n",
			smtp_strerror (smtp_errno (), buf, sizeof buf));
	else {
		/* Report on the success or otherwise of the mail transfer.
		 */
		if (debug) {
			status = smtp_message_transfer_status (message);
			printf ("%d %s", status->code,
				(status->text != NULL) ? status->text : "\n");
		}
		smtp_enumerate_recipients (message, print_recipient_status, NULL);
	}

	if (debug)
		fprintf(stderr, "Email sent successfully!\n");

	/* Free resources consumed by the program.
	 */
	smtp_destroy_session (session);
	return 0;
}
